#!/usr/bin/fades

import multiprocessing
import os
import struct
import sys
import threading
import time
import timeit
from multiprocessing import shared_memory

import numpy as np  # fades
import dask.array as da  # fades

n_cores = os.cpu_count() // 2
print("Trabajando con la mitado de los cores:", n_cores)


def bounds(parts, size):
    """Slide 12.

    Ejemplos:
        bounds(4, 23) -> [0, 6, 12, 18, 23]
        bounds(7, 25) -> [0, 4, 8, 12, 16, 19, 22, 25]
    """
    len_bigs = size % parts
    len_smalls = parts - len_bigs
    chunk_sizes = [size // parts + 1] * len_bigs + [size // parts] * len_smalls

    result = [0]
    for cs in chunk_sizes:
        result.append(result[-1] + cs)
    return result


def sumV(a, b, c, lim_inf, lim_sup, tid):
    """Slide 13."""
    for i in range(lim_inf, lim_sup):
        c[i] = a[i] + b[i]


def serial(size):
    """Cálculo serial."""
    a = [1] * size
    b = [3] * size
    c = [None] * size
    sumV(a, b, c, 0, size, None)
    return c


def threads(size):
    """Cálculo pseudo-paralelo (threads)."""
    a = [1] * size
    b = [3] * size
    bnd = bounds(n_cores, size)
    threads = []
    c = [None] * size
    for i in range(n_cores):
        t = threading.Thread(target=sumV, args=(a, b, c, bnd[i], bnd[i + 1], i))
        t.start()
        threads.append(t)
    for t in threads:
        t.join()
    return c


def processes_naive(size):
    """Cálculo paralelo real, versión inocente, **NO FUNCIONA**."""
    a = [1] * size
    b = [3] * size
    bnd = bounds(n_cores, size)
    processes = []
    c = [None] * size
    for i in range(n_cores):
        p = multiprocessing.Process(target=sumV, args=(a, b, c, bnd[i], bnd[i + 1], i))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()
    return c


def processes_array(size):
    """Cálculo paralelo real, con Array."""
    a = [1] * size
    b = [3] * size
    bnd = bounds(n_cores, size)
    processes = []
    c = multiprocessing.Array('B', size)

    for i in range(n_cores):
        p = multiprocessing.Process(target=sumV, args=(a, b, c, bnd[i], bnd[i + 1], i))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()
    return c


def processes_shareable_list(size):
    """Cálculo paralelo real, con ShareableList."""
    a = [1] * size
    b = [3] * size
    bnd = bounds(n_cores, size)
    processes = []
    s_c = shared_memory.ShareableList([0] * size)

    for i in range(n_cores):
        p = multiprocessing.Process(target=sumV, args=(a, b, s_c, bnd[i], bnd[i + 1], i))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()

    c = list(s_c)
    s_c.shm.close()
    s_c.shm.unlink()

    return c


def processes_raw_array(size):
    """Cálculo paralelo real, con RawArray."""
    a = [1] * size
    b = [3] * size
    bnd = bounds(n_cores, size)
    processes = []
    c = multiprocessing.RawArray('B', size)

    for i in range(n_cores):
        p = multiprocessing.Process(target=sumV, args=(a, b, c, bnd[i], bnd[i + 1], i))
        p.start()
        processes.append(p)

    for p in processes:
        p.join()
    return c


def processes_fork_v1(size):
    """Cálculo paralelo real, muy a mano, v1."""
    a = [1] * size
    b = [3] * size
    bnd = bounds(n_cores, size)
    type_size = struct.calcsize('B')
    c = [None] * size

    # shared memory to use from all the processes
    shm = shared_memory.SharedMemory(create=True, size=type_size * size)

    def child(child_number):
        """Work done in other process."""
        lim_inf = bnd[child_number]
        lim_sup = bnd[child_number + 1]

        # calculate the part of 'c' that we care
        sumV(a, b, c, lim_inf, lim_sup, child_number)

        # copy the part of c that we care to the shared memory location
        values = c[lim_inf:lim_sup]
        offset = type_size * lim_inf
        struct.pack_into(f"{len(values)}B", shm.buf, offset, *values)
        return SystemExit()

    for i in range(n_cores):
        pid = os.fork()
        if pid == 0:
            # child side
            raise child(i)

    # only the parent will reach here; wait for all children to continue
    for i in range(n_cores):
        os.wait()

    # get produced value and cleanup the shared memory
    c = struct.unpack(f"{size}B", shm.buf)
    shm.close()
    shm.unlink()

    return c


def processes_fork_v2(size):
    """Cálculo paralelo real, muy a mano, v2."""
    a = [1] * size
    b = [3] * size
    bnd = bounds(n_cores, size)
    type_size = struct.calcsize('B')

    # shared memory to use from all the processes
    shm = shared_memory.SharedMemory(create=True, size=type_size * size)

    def child(child_number):
        """Work done in other process."""
        lim_inf = bnd[child_number]
        lim_sup = bnd[child_number + 1]

        offset = lim_inf * type_size
        for idx in range(lim_inf, lim_sup):
            result = a[idx] + b[idx]
            shm.buf[offset] = result
            offset += type_size
        return SystemExit()

    for i in range(n_cores):
        pid = os.fork()
        if pid == 0:
            # child side
            raise child(i)

    # only the parent will reach here; wait for all children to continue
    for i in range(n_cores):
        os.wait()

    # get produced value and cleanup the shared memory
    c = struct.unpack(f"{size}B", shm.buf)
    shm.close()
    shm.unlink()

    return c


def numpy_simple(size):
    """Con NumPy, simple."""
    a = np.full(size, 1)
    b = np.full(size, 3)
    return a + b


def numpy_paralelo(size):
    """Con NumPy, procesos en paralelo."""
    a = np.full(size, 1)
    b = np.full(size, 3)
    bnd = bounds(n_cores, size)

    # shared memory to use from all the processes
    type_size = np.int8().size
    shm = shared_memory.SharedMemory(create=True, size=type_size * size)
    c = np.ndarray((size,), dtype=np.int8, buffer=shm.buf)

    def child(child_number):
        """Work done in other process."""
        lim_inf = bnd[child_number]
        lim_sup = bnd[child_number + 1]
        c[lim_inf:lim_sup] = a[lim_inf:lim_sup] + b[lim_inf:lim_sup]
        return SystemExit()

    for i in range(n_cores):
        pid = os.fork()
        if pid == 0:
            # child side
            raise child(i)

    # only the parent will reach here; wait for all children to continue
    for i in range(n_cores):
        os.wait()

    # get produced value (before!) and cleanup the shared memory (after!)
    c = np.array(c)
    shm.close()
    shm.unlink()

    return c


def dask(size):
    """Con Dask."""
    chunk_size = size // n_cores
    a = da.full(size, 1, chunk_size)
    b = da.full(size, 3, chunk_size)
    c = a + b
    return c.compute()


result_base = None


def test(func):
    global result_base

    size = 1000  # 50_007_007
    print(func.__doc__)
    t_start = time.time()
    c = func(size)
    print("    fin:", time.time() - t_start)
    if result_base is None:
        result_base = c
        print("    set base result")
    else:
        if result_base == list(c):
            print("    ok")
        else:
            print("    MAL:", repr(c))


def medir(func):
    # the function will be called 'inner_quant' times while GC is disabled (total measured time
    # needs to be divided by this, getting effectively the average); that will happen
    # 'outer_quant' times, and as timeit.Timer.repeat docstring states, it's better to get
    # the min of it
    inner_quant = 20
    outer_quant = 5

    # start with size of 1000 (doesn't make sense to measure less than that) and scales in a
    # way to get 3 measure points for each magnitud order
    base_size = 1000
    step = 10 ** (1 / 3)

    # measure until the result is higher than 10 seconds or the size reached 100 million
    time_limit = 3

    for idx in range(10000):
        size = int(base_size * step ** idx)
        if size > 100_000_000:
            break
        measures = timeit.repeat(lambda: func(size), number=inner_quant, repeat=outer_quant)
        useful = min(measures) / inner_quant
        report = (func.__name__, str(size), str(useful).replace(".", ","))
        print(";".join(report))
        if useful > time_limit:
            break


if len(sys.argv) != 2:
    print("Usar: vectores_paralelos.py {test|medir}")
    exit()

if sys.argv[1] == 'test':
    run = test
elif sys.argv[1] == 'medir':
    run = medir
else:
    print("Usar: vectores_paralelos.py {test|medir}")
    exit()


run(serial)  # base, secuencial
run(threads)  # hilos, obviamente el mismo proceso/core, costo de switch sin ninguna ganancia
# run(processes_naive)  # procesos, con 'c' compartido; acceso sincronizado, no escala
run(processes_array)  # procesos, con 'c' compartido; acceso sincronizado, no escala
run(processes_shareable_list)  # lista genérica, también es acceso sincronizado
run(processes_fork_v1)  # muy a mano, primer intento
run(processes_fork_v2)  # muy a mano, segundo intento
run(processes_raw_array)  # como Array, pero sin el costo de sincronización (lo podemos romper!)
run(numpy_simple)  # usando numpy directamente
run(numpy_paralelo)  # usando numpy en varios procesos
run(dask)  # y dask?
